/*
    Copyright (C) 2005 Sergey Koposov
  
  Author: Sergey Koposov, Sternberg Astronomical Institute
  Email: math@sai.msu.ru 
  http://lnfm1.sai.msu.ru/~math

  This file is part of GMATCH.

  GMATCH is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GMATCH is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GMATCH; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "wcs.h"
#include <math.h>
#include "common.h"

void our_wcs_init(struct wcsprm *our_wcs, int xsize, int ysize, double ra, double dec,
			                        double scale)
{

	memset(our_wcs,0,sizeof(our_wcs));
	our_wcs->flag = -1;   
	wcsini(1, 2, our_wcs);

	our_wcs->flag = 0; 
	our_wcs->naxis = 2;

	our_wcs->crpix[0] = (xsize + 1.) / 2.;
	our_wcs->crpix[1] = (ysize + 1.) / 2.;

	our_wcs->pc[0] = scale;           
	our_wcs->pc[1] = 0;    
	our_wcs->pc[2] = 0;
	our_wcs->pc[3] = scale;

	strcpy((our_wcs->cunit)[0], "deg");
	strcpy((our_wcs->cunit)[1], "deg");

	strcpy((our_wcs->ctype)[0], "RA---TAN");
	strcpy((our_wcs->ctype)[1], "DEC--TAN");

	strcpy(our_wcs->radesys, "ICRS");

	our_wcs->crval[0] = ra;           
	our_wcs->crval[1] = dec; 

	our_wcs->cdelt[0] = 1;           
	our_wcs->cdelt[1] = 1; 

	our_wcs->lonpole = 999;    
	our_wcs->latpole = 999.;  

	//our_wcs->equinox = 2000.;
	wcsset(our_wcs);

}

void init_arrays(int num, double **xy, double **phi,
				double **theta, double **imgcrd, int **stat)
{
	//*radec=(double *)malloc(2*num*sizeof(double));
	*xy = (double *)malloc(2 * num * sizeof(double));
	*phi = (double *)malloc(num * sizeof(double));
	*theta = (double *)malloc(num * sizeof(double));
	*imgcrd = (double *)malloc(2 * num * sizeof(double));
	*stat = (int *)malloc(num * sizeof(int));
} 

void init_arrays1(int num, double **radec, double **phi,
					double **theta, double **imgcrd, int **stat)
{
	*radec=(double *)malloc(2*num*sizeof(double));
	//*xy = (double *)malloc(2 * num * sizeof(double));
	*phi = (double *)malloc(num * sizeof(double));
	*theta = (double *)malloc(num * sizeof(double));
	*imgcrd = (double *)malloc(2 * num * sizeof(double));
	*stat = (int *)malloc(num * sizeof(int));
} 


static void convert(double ra_cen, double dec_cen, double *radec, int nstars,
					int xsize, int ysize, double scale, double **xy0)
{
	struct wcsprm our_wcs;
	double *xy, *phi, *theta, *imgcrd;
	int *stat;
	
	our_wcs_init(&our_wcs, xsize, ysize, ra_cen, dec_cen, scale);
	init_arrays(nstars, &xy, &phi, &theta, &imgcrd, &stat);
	*xy0 = xy;

	wcss2p(&our_wcs, nstars, 2, radec, phi, theta, imgcrd, xy, stat);
	free(phi);
	free(theta);
	free(imgcrd);
	free(stat);
	wcsfree(&our_wcs);
}

static void convert1(double *radec, int nstars, struct wcsprm *our_wcs, int xsize,
					int ysize, double **xy0)
{
	double *xy, *phi, *theta, *imgcrd;
	int *stat;
	
	init_arrays(nstars, &xy, &phi, &theta, &imgcrd, &stat);
	*xy0 = xy;

	wcss2p(our_wcs, nstars, 2, radec, phi, theta, imgcrd, xy, stat);
	free(phi);
	free(theta);
	free(imgcrd);
	free(stat);
}



/* Take into account that the array returned by this function should be freed */
double *read_and_wcs_convert(char *filename, double ra_cen, double dec_cen, double scale, int xsize, int ysize, int *nstars0)
{
	int max_nstars = 1000, i, nstars;
	FILE *fp;
	double *radec = malloc(2 * sizeof(double) * max_nstars);
	double *xy; 

	fp = fopen(filename, "r");

	if (!fp) 
	{
		perror(filename);
		free(radec);
		return 0;
	}

	for (i = 0; !feof(fp) && (i < max_nstars) && (i<*nstars0); )
	{
		char buf[101];
		fgets(buf,100,fp);
		if (2 == sscanf(buf, "%lf %lf", radec + 2 * i, radec + 2 * i + 1))
		{
			//fprintf(stdout, "%f %f\n",radec[2*i],radec[2*i+1]);
			i++;
		}
	}
	fclose(fp);
	nstars = i;
	
	convert(ra_cen, dec_cen, radec, nstars, xsize, ysize, scale, &xy);
	free(radec);
	*nstars0 = nstars;

	return xy;
}


/* Take into account that the array returned by this function should be freed */
double *read_and_wcs_convert1(char *filename, double ra_cen, double dec_cen,
							double scale, int xsize, int ysize,
							struct wcsprm *our_wcs, int *nstars0)
{
	int max_nstars = 1000, i, nstars;
	FILE *fp;
	double *radec = malloc(2 * sizeof(double) * max_nstars);
	double *xy, *last_radec;

	fp = fopen(filename, "r");

	if (!fp) 
	{
		perror(filename);
		free(radec);
		return 0;
	}

	for (i = 0; !feof(fp) && (i < max_nstars); )
	{
		char buf[101];
		fgets(buf,100,fp);
		if (2 == sscanf(buf, "%lf %lf", radec + 2 * i, radec + 2 * i + 1))
		{
			//fprintf(stdout, "%f %f\n",radec[2*i],radec[2*i+1]);
			i++;
		}
	}
	fclose(fp);
	
	nstars = i;
	//fprintf(stderr,"RRRR %d", nstars);

	convert1(radec, nstars, our_wcs, xsize, ysize, &xy);
	
	for (i = 0, last_radec = radec; 
			                (i < nstars) && ((last_radec - radec) < 2 * (*nstars0)); i++)
	{
		//fprintf(stderr, "RR %d %f %f\n", i, xy[2*i], xy[2*i+1]);
		if ((xy[2 * i] >= 1) && (xy[2 * i] <= xsize) &&
			  (xy[2 * i + 1] >= 1) && (xy[2 * i + 1] <= ysize))
		{
			//last_xy[0]=xy[2*i];
			//last_xy[1]=xy[2*i+1];
			last_radec[0] = radec[2 * i];
			last_radec[1] = radec[2 * i + 1];
			last_radec += 2;
		}
	}
	
	*nstars0 = (last_radec - radec) / 2;

	free(xy);
	
	convert(ra_cen, dec_cen, radec, *nstars0, xsize, ysize, scale, &xy);

	free(radec);

	return xy;
}




