/*
 *   Copyright (C) 2005 Sergey Koposov
 * 
 * Author: Sergey Koposov, Sternberg Astronomical Institute
 * Email: math@sai.msu.ru 
 * http://lnfm1.sai.msu.ru/~math

  This file is part of GMATCH.

  GMATCH is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GMATCH is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GMATCH; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
/* 
  Arguments: 
  1) Name of FITS file
  2) Name of file with the list of sextracted X and Y positions
  3) Name of file with the list of ra and dec positions around point of interest
  4) approximate RA of the center
  5) approximate DEC of the center
  6) number of stars which will be taken from the SEXlist
  7) number of stars which will be taken from the CONElist
  8) use_wcs flag
  
  Example:
  ./gmatch data.fits sex_list cone_list 21.14 -1.16 40 40 0 
*/

int main(int argc, char *argv[])
{

	char *fits_filename, *obs_list_filename, *std_list_filename;
	double crval[2], crpix[2], cd[4];  
	int nobsstars, nstdstars;
	char usewcs = 0; 
	
	if (argc != 9) 
	{
		fprintf(stderr,"Wrong arguments!!!\n");
		exit(1);
	}
	
	fits_filename = (char *)calloc(256, 1);
	obs_list_filename = (char *)calloc(256, 1);
	std_list_filename = (char *)calloc(256, 1);

	strncpy(fits_filename, argv[1],255);
	strncpy(obs_list_filename, argv[2],255);
	strncpy(std_list_filename, argv[3],255);

	gr_type ra_cen = atof(argv[4]);
	gr_type dec_cen = atof(argv[5]);

	nobsstars = atoi(argv[6]);
	nstdstars = atoi(argv[7]);
	usewcs = (atoi(argv[8])!=0);
	gr_type eps = 1;
	
	
	int status=run_groth(fits_filename, obs_list_filename, std_list_filename, ra_cen,
		dec_cen, eps, crpix, crval, cd, nobsstars, nstdstars, usewcs);
	
	if (status == 0) 
	{
		fix_wcs(fits_filename, cd, crpix, crval);
	}

	free(fits_filename);
	free(obs_list_filename);
	free(std_list_filename);
	
	return 0;
}
