/*
    Copyright (C) 2005 Sergey Koposov
  
  Author: Sergey Koposov, Sternberg Astronomical Institute
  Email: math@sai.msu.ru 
  http://lnfm1.sai.msu.ru/~math

  This file is part of GMATCH.

  GMATCH is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GMATCH is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GMATCH; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */


#include "common.h"
#include "config.h"
#include "common_wcs.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
int get_transform0(struct obs_list *olist, struct gr_xy *obs_xy,
				           struct gr_xy *std_xy, int nobstriangles, int nobsstars, char  *f,char * f2);

/* Take into account that the array returned by this function should be freed */
double *read_and_wcs_convert0(char *filename, int *nstars0)
{
	int max_nstars = 1000, i, nstars;
	FILE *fp;
	double *radec = malloc(2 * sizeof(double) * max_nstars);
	double *xy; 

	fp = fopen(filename, "r");

	if (!fp) 
	{
		perror(filename);
		free(radec);
		return 0;
	}

	for (i = 0; !feof(fp) && (i < max_nstars) && (i<*nstars0); )
	{
		if (2 == fscanf(fp, "%lf %lf", radec + 2 * i, radec + 2 * i + 1))
		{
			//fprintf(stdout, "%f %f\n",radec[2*i],radec[2*i+1]);
			i++;
		}
	}
	fclose(fp);
	nstars = i;
	
	xy=radec;
	*nstars0 = nstars;

	return xy;
}

void run_groth1(char *obs_list_filename,
				       char *std_list_filename,
				       gr_type eps, int nobsstars, int nstdstars, char *f1,char * f2)
{

	struct gr_xy *obs_xy, *std_xy;

	int i, nobstriangles, nstdtriangles,
			nstars;
	struct std_list *std_data;
	struct obs_list *obs_data;
	double *xy;
	void *ipix_ptr;
	FILE *fp;
	int status;
	
	obs_xy = malloc(sizeof(struct gr_xy) * nobsstars);
	
	nstars = nstdstars;

	xy = read_and_wcs_convert0(std_list_filename, &nstars);

	nstdstars = nstars;
	
	if (xy == 0)
	{
		free(obs_xy);
		fprintf(stderr,"Strange error!!!\n");
		return;
	}

	std_xy = malloc(sizeof(struct gr_xy) * nstdstars);
	
	for(i = 0; i < nstdstars; i++)
	{
		std_xy[i].x = xy[2 * i];
		std_xy[i].y = xy[2 * i + 1];
	}
	

	free(xy);

	fp = fopen(obs_list_filename, "r");
	if (!fp) 
	{
		perror(obs_list_filename);
		return;
	}

	for (i = 0; !feof(fp) && (i < nobsstars); )
	{
		if (2 == fscanf(fp, "%lf %lf", &(obs_xy[i].x), &(obs_xy[i].y)))
		{	
			i++;
		}
	}

	nobsstars = i;

	nobstriangles = (nobsstars * (nobsstars - 1) * (nobsstars - 2)) / 6;

	fclose(fp);

	process_obs_list (obs_xy, nobsstars, &obs_data, &nobstriangles, eps);
	process_std_list (std_xy, nstdstars, &std_data, &ipix_ptr, &nstdtriangles);

	status = match_lists(std_data, obs_data, ipix_ptr, nstdtriangles, nobstriangles);
	free(ipix_ptr);

	if (status != 0)
	{
		return;
	}
	
/*  for(i = 0;i < nobstriangles; i++)
	{
		if (obs_data[i].matched_ptr)
		{
			fprintf(stderr,"%d %d\n",obs_data[i].id, obs_data[i].matched_ptr->id);  
		}
	}
*/
	get_transform0(obs_data, obs_xy, std_xy, nobstriangles, nobsstars,f1,f2);


}


int get_transform0(struct obs_list *olist, struct gr_xy *obs_xy,
				           struct gr_xy *std_xy, int nobstriangles, int nobsstars, char *file1, char *file2)
{
	struct gr_xy *sxy, *oxy;
	gr_type *operim, *sperim, *ratio, *ratio1, mean_ratio, mean_ratio1, st_dev,
				  t, thresh, mf, mt, max, t1, tx, ty, tx1, ty1, txx, txy, txx1, txy1,
				  tyy, tyx1, tyy1, A, B, C, D, 
				  vall, valr;
	int nmatch, i, m0, m1, j, cur_n, n_final_triangles;
	double_and_char *ratio_and_flag;

	char *oorient, *sorient, orient;
	char *flag;
	struct three_ints *star_list;
	struct obs_list **optr;
	int *hist, nhist = 10, posl, posr, posl1, posr1, ret_val;
	gr_matrix m00;
	gr_vector v00;
	gr_matrix *m =&m00;
	gr_vector *v =&v00;
	memset(m,0,sizeof(gr_matrix));
	memset(v,0,sizeof(gr_vector));
	
	sxy = malloc(sizeof(struct gr_xy) * nobstriangles);
	oxy = malloc(sizeof(struct gr_xy) * nobstriangles);
	optr = malloc(sizeof(struct obs_list *) * nobstriangles);
	
	/* The list of perimeters of observed triangles */
	operim = malloc(sizeof(gr_type) * nobstriangles);

	/* The list of perimeters of standart triangles */
	sperim = malloc(sizeof(gr_type) * nobstriangles);

	/* The list of ratios of triangles */
	ratio = malloc(sizeof(gr_type) * nobstriangles);

	/* The list of orientations of standart triangles */
	sorient = malloc(nobstriangles);

	/* The list of orientations of observed triangles */
	oorient = malloc(nobstriangles);

	/* The space for the histogram used in mode determination*/
	hist = malloc((2 * nhist + 1) * sizeof(int));



	for(i = 0, mean_ratio = 0, nmatch = 0; i < nobstriangles; i++)
	{
		if (!(olist[i].matched_ptr)) continue;
		
		operim[nmatch] = olist[i].perim;
		oxy[nmatch] = obs_xy[olist[i].id]; 
		sperim[nmatch] = olist[i].matched_ptr->perim;
		sxy[nmatch] = std_xy[olist[i].matched_ptr->id];
		optr[nmatch] = olist + i;
		oorient[nmatch] = olist[i].orient;
		sorient[nmatch] = olist[i].matched_ptr->orient;
		ratio[nmatch] = 0.5 * log10(operim[nmatch] / sperim[nmatch]);
		mean_ratio += ratio[nmatch]; 
		
#ifdef GMATCH_DEBUG
		fprintf(stderr, "%d %f\n", nmatch, ratio[nmatch]);
#endif
		
		nmatch++;
	}
	
	mean_ratio /= nmatch;

	free(operim);
	free(sperim); 

	oxy = realloc(oxy, sizeof(struct gr_xy) * nmatch);
	sxy = realloc(sxy, sizeof(struct gr_xy) * nmatch);
	optr = realloc(optr, sizeof(struct obs_list *) * nmatch);
	
	oorient = realloc(oorient, nmatch);
	sorient = realloc(sorient, nmatch);
	ratio = realloc(ratio, sizeof(gr_type) * nmatch);
		
	/* this array will contain the sorted array of ratios of 
		 perimeters of triangles  */
	ratio1 = malloc(sizeof(gr_type) * nmatch);

	ratio_and_flag = malloc(sizeof(double_and_char) * nmatch);

	/* The flag will be used in false match removal 
		 (to flag the bad triangle matches)*/
	flag = calloc(nmatch , 1);

	
	/* That list will be used in final matching process */
	star_list = calloc(3 * nmatch, sizeof(struct three_ints));


	
	for(i = 0; i < nmatch; i++)
	{
		ratio1[i] = ratio[i];
		ratio_and_flag[i].ratio = ratio[i];
		ratio_and_flag[i].flag = (sorient[i]==oorient[i]);

	}
	qsort(ratio1, nmatch, sizeof(gr_type), gr_type_compare);
	qsort(ratio_and_flag, nmatch, sizeof(double_and_char), double_and_char_compare);

#ifdef GMATCH_DEBUG
	for(i = 0; i < nmatch; i++)
	{
		fprintf(stderr,"Y %f\n",ratio1[i]);
	}
#endif
	
	posl = 0; 
	posr = nmatch - 1;
	for(i = 0; (i < 8) && ((posr - posl) > 10) && ((ratio_and_flag[posr].ratio-ratio_and_flag[posl].ratio)>0.03); i++)
	{
		memset(hist, 0, (2 * nhist + 1) * sizeof(int));
		//mode_hist(ratio1 + posl, posr - posl + 1, hist, nhist, &posl1, &posr1);
		mode_hist1(ratio_and_flag + posl, posr - posl + 1, hist, nhist, &posl1, &posr1);

		posr = posr1 + posl;
		posl = posl1 + posl;

#ifdef GMATCH_DEBUG
		fprintf(stderr, "T %f %f %d %d\n",ratio1[posl],ratio1[posr], posl, posr);
#endif
	}
	vall=ratio1[posl];
	valr=ratio1[posr];

	for(i = posl, mean_ratio = 0; i <= posr; i++)
	{
		mean_ratio+=ratio1[i];
	}
	mean_ratio /= (posr - posl + 1);

	for(i = 0, st_dev = 0, m0 = 0, m1 = 0; i < nmatch; i++)
	{
		if ((ratio[i] < vall) || (ratio[i] > valr)) 
		{
			continue;
		}
		t = (ratio[i] - mean_ratio);
		st_dev += t * t;
		if (sorient[i] != oorient[i]) 
		{
			m0++;
		}
		else
		{
			m1++;
		}
	}
	
	st_dev = sqrt(st_dev / (posr-posl+1));
	
	
	/* We perform 10 iterations to reject the false matches */
	for(j = 0; j < 10; j++)
	{

#ifdef GMATCH_DEBUG    
		fprintf(stderr,"U %d %d %d %f %f\n", nmatch, m0, m1, mean_ratio, st_dev);
#endif

		mt = ((m1 - m0) > 0) ? (m1 - m0) : (m0 - m1);
		mf = m1 + m0 - mt;
		if (mf > mt)
		{
			thresh = 1;
		}
		else
		{
			if (0.1 * mt > mf)
			{
				thresh = 3;
			}
			else
			{
				thresh = 2;
			}
		}
		
		for(i = 0, mean_ratio1 = 0, cur_n = 0; i < nmatch; i++)
		{
			if (flag[i])
			{
				continue;
			}
			
			if (fabs(ratio[i] - mean_ratio) > thresh * st_dev)
			{
				flag[i] = 1;
				continue;
			}
			
			mean_ratio1 += ratio[i];
			cur_n++;      
		}
		
		mean_ratio = mean_ratio1 / cur_n;
		
		for(i = 0, m0 = 0, m1 = 0, st_dev = 0; i < nmatch; i++)
		{
			if (flag[i])
			{
				continue;
			}

			t = (ratio[i] - mean_ratio);
			st_dev += t * t;
			if (sorient[i] != oorient[i])
			{
				m0++;
			}
			else
			{
				m1++;
			}
		}
		st_dev = sqrt(st_dev / cur_n); 
	}
	
	if (m1 > m0)
	{
		for(i = 0; i < nmatch; i++)
		{
			if (sorient[i] != oorient[i])
			{
				flag[i] = 1;
			}
		}
		orient = 1;
	}
	else
	{
		for(i = 0; i < nmatch; i++)
		{
			if (sorient[i] == oorient[i])
			{
				flag[i] = 1;
			}
		}  
		orient = 0;
	}
	
	
	
	for(i = 0, n_final_triangles = 0; i < nmatch; i++)
	{
		if (flag[i])
		{
			continue;
		}
		/*
		fprintf(stderr,"%d %d %d %f %f %f %f\n",i,optr[i]->id, optr[i]->matched_ptr->id,
				    obs_xy[optr[i]->id].x, 
				    obs_xy[optr[i]->id].y, std_xy[optr[i]->matched_ptr->id].x, 
				    std_xy[optr[i]->matched_ptr->id].y);
		fprintf(stderr,"%d %d %d %f %f %f %f\n",i,optr[i]->id3, optr[i]->matched_ptr->id3,
				    obs_xy[optr[i]->id3].x, 
				    obs_xy[optr[i]->id3].y, std_xy[optr[i]->matched_ptr->id3].x, 
				    std_xy[optr[i]->matched_ptr->id3].y);
		fprintf(stderr,"%d %d %d %f %f %f %f\n",i,optr[i]->id2, optr[i]->matched_ptr->id2,
				    obs_xy[optr[i]->id2].x, 
				    obs_xy[optr[i]->id2].y, std_xy[optr[i]->matched_ptr->id2].x, 
				    std_xy[optr[i]->matched_ptr->id2].y);
		*/

		star_list[3*n_final_triangles].l2 = optr[i]->id;
		star_list[3*n_final_triangles].l3 = optr[i]->matched_ptr->id;
		star_list[3*n_final_triangles+1].l2 = optr[i]->id2;
		star_list[3*n_final_triangles+1].l3 = optr[i]->matched_ptr->id2;
		star_list[3*n_final_triangles+2].l2 = optr[i]->id3;
		star_list[3*n_final_triangles+2].l3 = optr[i]->matched_ptr->id3;
		n_final_triangles++;
	}

	/*
	for(i = 0; i < 3 * n_final_triangles; i++)
	{
		fprintf(stderr,"X %d %d %d %d\n",i,star_list[i].l1,star_list[i].l2,star_list[i].l3);
	}
	*/

	qsort(star_list, 3 * n_final_triangles, sizeof(struct three_ints), &three_ints_compare1);

	/*
	for(i = 0; i < 3 * n_final_triangles; i++)
	{
		fprintf(stderr,"A %d %d %d\n",star_list[i].l1,star_list[i].l2,star_list[i].l3);
	}
	*/

	for(i = 0, m0 = 0; i < 3 * n_final_triangles; i++)
	{
		if (m0 == 0)
		{
			m0 = 1;
			m1 = i;
			star_list[i].l1+=1;
		}
		else 
		{
			if (three_ints_compare1(star_list + m1, star_list + i) == 0)
			{
				star_list[m1].l1 += 1;
				continue;
			
			}
			else 
			{
				star_list[i].l1 += 1;
				m1 = i;
				m0 = 1;
			}
		}
	}

	/*
	for(i = 0;i<3 * n_final_triangles; i++)
	{
		fprintf(stderr,"B %d %d %d\n",star_list[i].l1,star_list[i].l2,star_list[i].l3);
	}
	*/
/*
	if (n_final_triangles == nmatch) 
	{
		fprintf(stderr,"ERROR: Triangles do not match!!!\n");
		return -1;
	}
	*/
	qsort(star_list, 3 * n_final_triangles, sizeof(struct three_ints), &three_ints_compare);
	

#ifdef GMATCH_DEBUG  
	fprintf(stderr,"\n");
	for(i = 0; i < 3 * n_final_triangles; i++)
	{
		fprintf(stderr,"C %d %d %d\n", star_list[i].l1, star_list[i].l2, star_list[i].l3);
	}
#endif  

	max = star_list[0].l1;

	if (max < 1) 
	{
		fprintf(stderr,"ERROR: Triangles do not match!!!\n");
		ret_val = -1;
		goto GET_TRANSFORM_END;
	}

 FILE *f1 = fopen(file1,"w+");
	for (i = 0; i < 3 * n_final_triangles && (star_list[i].l1 >= (max / 2)); i++)
	{
		
		fprintf(f1,"%d %d\n",star_list[i].l2, star_list[i].l3);
#ifdef GMATCH_DEBUG
		fprintf(stderr,"W %d %d %d %f %f %f %f\n",star_list[i].l1, star_list[i].l2,
						star_list[i].l3, obs_xy[star_list[i].l2].x, 
						obs_xy[star_list[i].l2].y, std_xy[star_list[i].l3].x, 
						std_xy[star_list[i].l3].y);
#endif
		
		t1 += 1;  
		tx   += obs_xy[star_list[i].l2].x;
		ty   += obs_xy[star_list[i].l2].y;
		tx1  += std_xy[star_list[i].l3].x;
		ty1  += std_xy[star_list[i].l3].y;
		txx  += obs_xy[star_list[i].l2].x * obs_xy[star_list[i].l2].x;
		txy += obs_xy[star_list[i].l2].x * obs_xy[star_list[i].l2].y;
		txx1 += obs_xy[star_list[i].l2].x * std_xy[star_list[i].l3].x;
		txy1 += obs_xy[star_list[i].l2].x * std_xy[star_list[i].l3].y;
		tyy  += obs_xy[star_list[i].l2].y * obs_xy[star_list[i].l2].y;
		tyx1 += obs_xy[star_list[i].l2].y * std_xy[star_list[i].l3].x;
		tyy1 += obs_xy[star_list[i].l2].y * std_xy[star_list[i].l3].y;


	}
 fclose(f1);


	if (orient == 0) 
	{
		get_regression2_1(t1, tx, ty, tx1, ty1, txx, txx1, txy1, tyy, tyx1,
				              tyy1, &A, &B, &C, &D);
	}
	else
	{
		get_regression2  (t1, tx, ty, tx1, ty1, txx, txx1, txy1, tyy, tyx1,
				              tyy1, &A, &B, &C, &D);
	}

	m->t[0] = A;
	m->t[1] = B;
	if (orient == 0)
	{
		m->b[0] = B;
		m->b[1] = -A;

	}
	else
	{
		m->b[0] = -B;
		m->b[1] = A;
	}
	v->t[0] = C;
	v->b[0] = D;


/*               !!  TODO  !!

 Here I should insert the code to compute the RMS of transformation 
 
 */
 
 FILE *f2 = fopen(file2,"w+");
	fprintf(f2,"%f %f %f %f %f %f\n",m->t[0], m->t[1], m->b[0], m->b[1],
				  v->t[0], v->b[0]);
fclose(f2);
//  get_regression(t1, tx, ty, tx1, txx, txx1, txy, tyx1, tyy, &A, &B, &C);
//  get_regression(t1, tx, ty, ty1, txx, txy1, txy, tyy1, tyy, &A, &B, &C);

//  fprintf(stderr,"%f %f %f %f\n", A, B, C, D);

	
	ret_val=0;
	
	GET_TRANSFORM_END:
	
	free(oxy);
	free(sxy);
	free(optr);
	free(oorient);
	free(sorient);
	free(ratio);
	free(ratio1);
	free(flag);
	free(hist);
	free(star_list);

	return ret_val;

}


/* 1st argument is the list with the x,y's of the observed list 
 * 2nd argument is the list with the x,y's of the standart list
 * 3rd argument is the output file which will contain the matched pairs
 	of points ids
 * 4th argument is the output file with the computed matrix
 */
int main(int argc, char* argv[])
{
	double eps=1.;
	int nobsstars=1000;
	int nstdstars=1000;
	if (argc!=5)
	{
		fprintf(stderr,"Wrong args!");
		exit(1);
	}
	char *obsname = argv[1];
	char *stdname = argv[2];
	char *outname1= argv[3];
	char *outname2 = argv[4];

	run_groth1(obsname,stdname,  eps, nobsstars, nstdstars, outname1, outname2);
	return 0;
}

